package com.elphel.imagej.orthomosaic;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;

import com.elphel.imagej.common.ShowDoubleFloatArrays;
import com.elphel.imagej.tileprocessor.Correlation2d;
import com.elphel.imagej.tileprocessor.TileNeibs;

import ij.ImagePlus;
import ij.gui.PointRoi;

public class ObjectLocation {
	String name;
	double [] xy_meters;
	double [] xy_pixels;
	
	public ObjectLocation (
			String name,
			double [] xy_meters) {
		this.name = name;
		this.xy_meters = xy_meters;
	}
	public String getName() {
		return name;
	}
	public double [] getMetric() {
		return xy_meters;
	}
	public int [] getPixels() {
		return new int[] {
				(int) Math.round(xy_pixels[0]),
				(int) Math.round(xy_pixels[1])};
	};
	
	public void setPixels(double [] xy) {
		xy_pixels = xy;
	}
	
	public double [] extractObjectImage(
			OrthoMapsCollection maps_collection,
			int size) {
		int nmap =maps_collection.getIndex(getName());
		OrthoMap [] ortho_maps=maps_collection.getMaps();
		final int width =     ortho_maps[nmap].getImageData().width;
		final int height =    ortho_maps[nmap].getImageData().height;
		final float [] src_img =  ortho_maps[nmap].getImageData().data;
		double [] dcrop = new double [size*size];
		Arrays.fill(dcrop, Double.NaN);
		int hsize = size/2;
		int [] xy = getPixels();
		boolean has_NaN=false;		
		for (int y = 0; y < size; y++) {
			int src_y = y - hsize + xy[1];
			if ((src_y >= 0) && (y < height)) {
				for (int x = 0; x < size; x++) {
					int src_x = x - hsize + xy[0];
					if ((src_x >= 0) && (x < width)) {
						double d = src_img[src_x + src_y * width];
						dcrop[x + size * y] = d;
						has_NaN |= Double.isNaN(d);
					} else {
						has_NaN=true;
					}
				}
			} else {
				has_NaN=true;
			}
		}
		if (has_NaN) {
			TileNeibs tn =  new TileNeibs(size,size);
			OrthoMap.fillNaNs(
					dcrop, // double [] data,
					tn, // TileNeibs tn,
					3); // int min_neibs)
		}
		return dcrop;
	}
	
	
	public static double [] getPatternCenter(
			double [] data_in,
			double [] pattern,
			double [] kernel,
			int       zoomout,
			double    phaseCoeff,
			double    min_corr,
			int       radius_search,
			int       radius_centroid,
			String    dbg_prefix,
			int       debugLevel){
		int refine = 1;
		int extr_size =    (int) Math.sqrt(data_in.length);
		int corr_size = 1;
		for (; corr_size <= extr_size; corr_size*=2);
		corr_size /=2;
		double [] data = data_in;
		if (corr_size < extr_size) {
			data = new double [corr_size*corr_size];
			int offs = extr_size/2 - corr_size/2;
			for (int i = 0; i < corr_size; i++) {
				System.arraycopy(
						data_in,
						offs * (extr_size + 1) + i * extr_size,
						data,
						i * corr_size,
						corr_size);
			}
		}
		
		int pattern_size = (int) Math.sqrt(pattern.length);
		/*
		// Already done while extracting
		boolean has_NaN=false;
		for (int i = 0; i < data.length; i++) {
			if (Double.isNaN(data[i])) {
				has_NaN=true;
				break;
			}
		}
		if (has_NaN) {
			TileNeibs tn =  new TileNeibs(corr_size,corr_size);
			OrthoMap.fillNaNs(
					data, // double [] data,
					tn, // TileNeibs tn,
					3); // int min_neibs)
		}
        */
		if (radius_search > (corr_size/2 - 1)) {
			radius_search = corr_size/2 - 1;
			System.out.println("getPatterCenter(): limiting radius_search to "+radius_search);
		}
		if (radius_centroid > radius_search) {
			radius_centroid = radius_search;
			System.out.println("getPatterCenter(): limiting radius_centroid to "+radius_centroid);
		}
		double [] corr_pattern = OrthoMap.patternZoomCropPad(
				pattern,      // double [] pattern,
				pattern_size, // int       pattern_size,
				corr_size,    // int       size,
				zoomout,      // int       zoomout,
				false); // true);        // out_normalize); // boolean   normalize) 
		if (kernel != null) {
			corr_pattern = OrthoMap.convolveWithKernel(
					corr_pattern,   // final double [] data,
					kernel,           // final double [] kernel,
					corr_size);       // final int width)
		}
		double [] corr_out= OrthoMap.correlateWithPattern(
				data,             // final double [] data,
				corr_size,        // final int       width,
				corr_size,        // final int       psize,      // power of 2, such as 64
				corr_pattern,     // final double [] pattern,    // [psize*psize]
				false,            // final boolean   convolve,   // convolve, not correlate
				phaseCoeff,       // final double    phaseCoeff,
				debugLevel);      // final int       debugLevel)
		int data_width = 2 *radius_search + 1;
		int offs = corr_size/2 - radius_search;
		
		double [] data_max = new double [data_width*data_width];
		for (int i = 0; i < data_width; i++) {
			System.arraycopy(
					corr_out,
					offs * (corr_size + 1) + i * corr_size,
					data_max,
					i * data_width,
					data_width);
		}
		// TODO: use min_corr and clusterize! Or use fixed fraction (50%) of the maximum ?
		double [] xys = Correlation2d.getMaxXYCm( // last, average
				data_max,        // corrs.length-1], // double [] data,
				data_width,      // int       data_width,      //  = 2 * transform_size - 1;
				radius_centroid, // double    radius, // 0 - all same weight, > 0 cosine(PI/2*sqrt(dx^2+dy^2)/rad)
				refine,          // int       refine, //  re-center window around new maximum. 0 -no refines (single-pass)
				null,            // boolean [] fpn_mask,
				false,           // boolean    ignore_border, // only if fpn_mask != null - ignore tile if maximum touches fpn_mask
				false);          // boolean   debug)

		
		if (dbg_prefix != null) {
			String []   dbg_titles = {"corr", "orig", "pattern"};
			double [][] dbg_img = {corr_out, data, corr_pattern};
			ImagePlus imp_corrs = ShowDoubleFloatArrays.makeArrays(
					dbg_img,
					corr_size,
					corr_size,
					dbg_prefix+"_pc"+phaseCoeff+"_rad"+radius_centroid,				
					dbg_titles); // test_titles,
			PointRoi roi = new PointRoi();
			roi.setOptions("label");
			roi.addPoint(xys[0]+corr_size/2, xys[1]+corr_size/2); // ,1);
			imp_corrs.setRoi(roi);
			imp_corrs.show();
		}
		return xys;
	}

	public static double [] getRadialPattern(
			final double [] data,
			double []  xy_offs,
			int        corr_size, 
			double     lim_rad,     // outside all 0
			double     trans_width, // lim_rad-trans_width - start reducing
			double     frac_outliers,
			double     reversal_rad,
			boolean [] good,       // null or same size as data
			int        debugLevel) {
		if (good != null) {
			Arrays.fill(good, false);
		}
		double min_abs_weight = 10.0; // Do not touch very center?
		int extr_size = (int) Math.sqrt(data.length);
		double [] rad_patt = new double [corr_size*corr_size];
		double xc = extr_size/2+xy_offs[0]; 
		double yc = extr_size/2+xy_offs[1]; 
		int size_1d = (int) Math.ceil(lim_rad);
		double lr2 = size_1d * size_1d; 
		double [] sw =  new double [size_1d + 2];
		double [] swd = new double [sw.length];
		double [] swd2 = new double [sw.length];
//		boolean [] good = new boolean [data.length];
		double [] distance = new double [data.length];
		ArrayList<ArrayList<Integer>> rad_lists = new ArrayList<ArrayList<Integer>>();
		for (int i = 0; i < sw.length; i++) {
			rad_lists.add(new ArrayList<Integer>());
		}
		for (int y = 0; y < extr_size; y++) {
			double dy = y - yc;
			double y2 = dy*dy;
			if (y2 <= lr2) {
				for (int x = 0; x < extr_size; x++) {
					double dx = x - xc;
					double r2 = y2 + dx*dx;
					if (r2 <= lr2) {
						int indx = x + extr_size * y;
						double r = Math.sqrt(r2);
						int ir = (int) Math.floor(r);
						distance[indx] = r;
						rad_lists.get(ir).add(indx);
						rad_lists.get(ir+1).add(indx);
						double w1 = r-ir;
						double w0 = 1.0 - w1;
						double d = data[indx];
						sw[ir] +=   w0;
						swd[ir] +=  w0 * d;
						swd2[ir] += w0 * d * d;
						sw[ir+1] +=  w1;
						swd[ir+1] += w1 * d;
						swd2[ir+1] += w1 * d * d;
						if (good != null) {
							good[indx] = true;
						}
					}
				}
			}
		}
		//size_1d
		for (int i = 0; i < sw.length; i++) {
			if (sw[i] > 0) {
				swd[i] /=  sw[i];
				swd2[i] /= sw[i];
			}
		}
		double [] sw0 = sw.clone(); // to watch how many were removed (frac_outliers)
		// now sort each list and later remove from either
		for (ArrayList<Integer> rad_list:rad_lists) {
			Collections.sort(rad_list, new Comparator<Integer>() {
				@Override
				public int compare(Integer lhs, Integer rhs) {
					double rhsd = data[rhs];
					double lhsd = data[lhs];
					return (rhsd > lhsd) ? -1 : (rhsd < lhsd) ? 1 : 0;
				}
			});
		}
		
		// min_weight = 10.0
		// Starting from very center remove fartherst (first or last) pixel,
		// until remaining weight drops below min_weight or sw0[]*(1-frac_outliers).
		// weight of the previous radius may fall also, but likely not too much
		for (int nr = 1; nr < sw.length; nr++) {
			double min_weight = Math.max(min_abs_weight, sw0[nr] * (1-frac_outliers));
			ArrayList<Integer> rad_list = rad_lists.get(nr);
			ArrayList<Integer> rad_list_prev = (nr > 0)? rad_lists.get(nr-1): null;
			ArrayList<Integer> rad_list_next = (nr < (sw.length -1))? rad_lists.get(nr+1): null;
			while (sw[nr] > min_weight) { // will break trying to remove last
				if (rad_list.size()==0) {
					System.out.println("BUG! - empty list for nr="+nr);
					break;
				}
				int indx_first = rad_list.get(0);
				int indx_last = rad_list.get(rad_list.size()-1);
				boolean remove_first = Math.abs(data[indx_first] - swd[nr]) > Math.abs(data[indx_last] - swd[nr]);
				int indx_worst = remove_first? indx_first : indx_last;
				double d = data[indx_worst];
				double dist = distance[indx_worst];
				boolean other_next = dist >= nr;
				double removed_w = other_next ? ( 1 - (dist - nr)) : (1 + (dist - nr));
				double new_w = sw[nr] - removed_w;
				if (new_w < min_weight) {
					break;
				}
				swd[nr] =  (swd[nr] *  sw[nr] - removed_w *  d) / new_w;
				swd2[nr] = (swd2[nr] * sw[nr] - removed_w *  d * d) / new_w;
				sw[nr] = new_w;
				rad_list.remove(remove_first ? 0 : (rad_list.size()-1));
				if (good != null) {
					good[indx_worst] = false; // seems not used - just for debug/display
				}
				// now remove from previous/next list
				if (dist == nr) {
					break;
				}
				int nr_other = other_next? (nr+1): (nr-1);
				ArrayList<Integer> rad_list_other = other_next? rad_list_next : rad_list_prev;
				// check for consistency:
				int indx_in_other = rad_list_other.indexOf(indx_worst);
				if (indx_in_other < 0) {
					System.out.println("BUG in getRadialPattern(), pixel "+indx_worst+" from nr="+nr+" is missing in "+nr_other);
					break;
				}
				double removed_w_other = 1.0 - removed_w;
				double new_w_other = sw[nr_other] - removed_w_other; // no check for remaining weight here
				swd[nr_other] =  (swd[nr_other] *  sw[nr_other] - removed_w_other *  d) / new_w_other;
				swd2[nr_other] = (swd2[nr_other] * sw[nr_other] - removed_w_other *  d * d) / new_w_other;
				sw[nr_other]= new_w_other;
				rad_list_other.remove(indx_in_other);
			}
		}
		
		int lr = size_1d;
		if (reversal_rad > 0) {
			int irr = (int) reversal_rad;
			boolean is_growing = swd[irr+1] > swd[irr];
			for (lr = irr+1; lr <= size_1d; lr++) {
				if (is_growing) {
					if (swd[lr+1] < swd[lr]) {
						break;
					}
				} else {
					if (swd[lr+1] > swd[lr]) {
						break;
					}
				}
			}
		}
		
		
		for (int i = 0; i < lr; i++) {
			swd[i] -= swd[lr];
		}
		for (int i = lr; i <swd.length; i++) {
			swd[i] = 0;
		}
		
		// TODO: add outlier removal here
		// apply window
		if (trans_width > 0) {
			for (int i = (int) (lim_rad-trans_width); i < sw.length; i++) {
				if (i < lim_rad) {
					swd[i] *= 0.5*(1.0 + Math.cos(Math.PI*(i - (lim_rad-trans_width))/trans_width));
				} else {
					swd[i] = 0;
				}
			}
		}
		// generate output pattern
		int oc = corr_size/2;
		for (int y = 0; y < corr_size; y++) {
			double dy = y - oc;
			double y2 = dy*dy;
			if (y2 <= lr2) {
				for (int x = 0; x < corr_size; x++) {
					double dx = x - oc;
					double r2 = y2 + dx*dx;
					if (r2 <= lr2) {
						double r = Math.sqrt(r2);
						int ir = (int) Math.floor(r);
						double w1 = r-ir;
						double w0 = 1.0 - w1;
						rad_patt[x + corr_size * y] = w0 * swd[ir] + w1 * swd[ir + 1];
					}
				}
			}
		}
		return rad_patt;
	}
	
	
	
	
}
